# %%
# Likely needed imports
import numpy as np




# %%
# Function to return the big numbers list
def create_big_numbers_list():
    big_numbers_list = [i for i in range(1000)]
    big_numbers_list[90] = 900
    big_numbers_list[345] = 200
    big_numbers_list[505] = 606
    big_numbers_list[786] = 323
    return big_numbers_list




# %%
# Function to return the strange phobias list
def create_strange_phobias_list():
    strange_phobias_list = [
        'Arachibutyrophobia'
        , 'Nomophobia'
        , 'Arithmophobia'
        , 'Plutophobia'
        , 'Xanthophobia'
        , 'Ablutophobia'
        , 'Octophobia'
        , 'Optophobia'
        , 'Globophobia'
        , 'Hippopotomonstrosesquippedaliophobia'
        , 'Ephebiphobia'
        , 'Omphalophobia'
        , 'Linonophobia'
        , 'Pogonophobia'
        , 'Chaetophobia'
        , 'Vestiphobia'
        , 'Ergophobia'
        , 'Decidophobia'
        , 'Eisoptrophobia'
        , 'Deipnophobia'
        , 'Phobophobia'
        , 'Nomophobia'
        ]
    return strange_phobias_list




# %%
# The main set of function checks separated by number
# Check the sum of the big numbers list
def list_check_1(x):
    if x == sum(create_big_numbers_list()):
        print('True')
    else:
        print('False')

def list_check_2(x):
    if x == min(create_big_numbers_list()):
        print('True')
    else:
        print('False')

def list_check_3(x):
    if x == max(create_big_numbers_list()):
        print('True')
    else:
        print('False')

def list_check_4(x):
    if x == len(create_big_numbers_list()):
        print('True')
    else:
        print('False')

def list_check_5(x):
    if x == sum(create_big_numbers_list())/len(create_big_numbers_list()):
        print('True')
    else:
        print('False')

def list_check_6(x):
    if x == create_strange_phobias_list()[0]:
        print('True')
    else:
        print('False')

def list_check_7(x):
    if x == create_strange_phobias_list().index('Globophobia'):
        print('True')
    else:
        print('False')

def list_check_8(x):
    if x == create_big_numbers_list()[0:10]:
        print('True')
    else:
        print('False')

def list_check_9(x):
    if x == ['Ablutophobia', 'Arachibutyrophobia', 'Arithmophobia']:
        print('True')
    else:
        print('False')

def list_check_10(x):
    if x == [0, 2, 4]:
        print('True')
    else:
        print('False')

def list_check_11(x):
    if x == 'Nomophobia':
        print('True')
    else:
        print('False')

def list_check_12(x):
    if x == False:
        print('Correct!')
    else:
        print('Nope!')




# %%
# Function to check the bonus
def list_check_bonus(bad_idxs, bad_values):
    bad_idxs_list = [90, 345, 505, 786]
    bad_values_list = [900, 200, 606, 323]
    if (bad_idxs == bad_idxs_list) and (bad_values == bad_values_list):
        print('True')
    else:
        print('False')


# %%
import numpy as np
import pandas as pd
pd.set_option('display.max_columns', 500)
import string

# %%
df = pd.read_csv('twitter_human_bots_dataset.csv', low_memory=False)
df.head()

# %%
df['created_at'] = pd.to_datetime(df['created_at'])
df['hour_created'] = pd.to_datetime(df['created_at']).dt.hour
df['avg_favourites_per_day'] = df.favourites_count / df.account_age_days
df['avg_statuses_per_day'] = df.statuses_count / df.account_age_days
df['avg_friends_per_day'] = df.friends_count / df.account_age_days
df['avg_followers_per_day'] = df.followers_count / df.account_age_days
df['has_profile_background'] = df.profile_background_image_url.isna()
df['has_profile_description'] = df.description.isna()
df['has_profile_image'] = df.profile_image_url.isna()
df['screen_name_length'] = df.screen_name.apply(len)
df['screen_name_capital_count'] = df.screen_name.fillna('').apply(lambda x: sum(1 for c in x if c.isupper()))
df['screen_name_punct_count'] = df.screen_name.fillna('').apply(lambda x: sum([1 for c in x if c in string.punctuation]))
df['description_length'] = df.description.fillna('').apply(len)
df['description_capital_count'] = df.description.fillna('').apply(lambda x: sum(1 for c in x if c.isupper()))
df['description_punct_count'] = df.description.fillna('').apply(lambda x: sum([1 for c in x if c in string.punctuation]))
df['description_at_count'] = df.description.fillna('').apply(lambda x: sum([1 for c in x if c == '@']))
df['description_link_count'] = df.description.fillna('').apply(lambda x: x.count('http'))
df['description_hash_count'] = df.description.fillna('').apply(lambda x: x.count('#'))
df.drop(columns=['Unnamed: 0', 'description', 'profile_background_image_url', 'profile_image_url', 'screen_name'], inplace=True)
df.head()

# %%
df.to_csv('twitter_bot_detection.csv', index=False)
# %%



# %%
import numpy as np
import pandas as pd

# %%
df = pd.read_csv('Managerial_and_Decision_Economics_2013_Video_Games_Dataset.csv'
                 , encoding = 'ISO-8859-1')
df.head()

# %%
l_cols = [
    'Console'
    , 'Title'
    , 'US Sales (millions)'
    , 'YearReleased'
    , 'Publisher'
    , 'Genre'
    , 'Sequel'
    , 'Re-release'
    , 'Usedprice'
    , 'Review Score'
    , 'RatingE'
    , 'RatingT'
    , 'RatingM'
    , 'MaxPlayers'
    , 'Online'
    , 'Handheld'
    , 'LtdEdition'
    , 'Multiplatform'
]
df = df[l_cols]

# %%
l_common_genres = df.Genre.value_counts()[0:5].index.to_list()
df['Is_Common_Genre'] = df.Genre.apply(lambda x: True if x in l_common_genres else False)


# %%
s_log_sales_ign = (np.log(df['US Sales (millions)']) + 80)
s_ign = (s_log_sales_ign + (np.random.choice([1,-1], size=len(df)) * s_log_sales_ign * np.random.normal(loc=0.05, scale=0.025, size=len(df)))).round()
df['IGN Review Score'] = s_ign
s_ign.describe()


# %%
df.to_csv('video_game_sales.csv', index=False)

# %%



# %%
# Function imports
import pandas as pd
import numpy as np
import string
import re
from fuzzywuzzy import fuzz
from fuzzywuzzy import process




# %%
# Function to clean the datasets
def clean_ratings(df_ratings_input, df_books_input, n_ratings_to_keep_books=25, n_ratings_to_keep_users=25, use_implicit='no'):

    df_ratings = df_ratings_input.copy()
    df_books = df_books_input.copy()

    # Get unique ISBNs in both dfs
    a_ratings_ISBN = df_ratings.ISBN.unique()
    a_books_ISBN = df_books.ISBN.unique()
    # And filter ratings down to only ISBNS in the books df
    df_ratings = df_ratings.query('ISBN in @a_books_ISBN').reset_index(drop=True)

    # Compress a couple authors into the same author
    d_author_replace = {
        'Dean R. Koontz': 'Dean Koontz'
        }
    df_books['Book-Author'] = df_books['Book-Author'].replace(d_author_replace)

    # Now deal with there being many ISBNs for the same book in the eeasiest way possible
    # Group by (cleaned) title author combo, to create a parent isbn
    # Then remap the ratings isbns to that new parent
    # The raw list here is kind of dumb, but effective for the popular authors and I got in too deep
    l_replace_general = [
        "'"
        , 'Uncut Edition', 'Oprahs Book Club'
        , 'A Stephanie Plum Novel', 'Stephanie Plum Novels', 'Janet Evanovichs Full Series'
        , 'Lives of the Mayfair Witches', 'Or There and Back Again', ': The Enchanting Prelude to The Lord of the Rings'
        , 'Grafton, Sue. Kinsey Millhone Mysteries.'
        , 'Kinsey Millhone Mysteries', 'A Kinsey Millhone Mystery', 'Tom Clancys Op Center'
        , 'Tom Clancys Power Plays', 'Dollanger Saga', ': A Memoir', 'An Anita Blake, Vampire Hunter Novel'
        , 'An Alex Delaware Novel', ': The Nice and Accurate Prophecies of Agnes Nutter, Witch'
        , ': A Fairy Story'
        ]
    l_replace_general = [item.lower() for item in l_replace_general]
    l_replace_general.sort(key=len, reverse=True)
    l_replace_other = [
        "'"
        , 'the', 'and', '&', 'amp'
        , 'Paperback', 'Hardcover', 'Paper', 'A Novel', 'Anniversary edition', 'A memoir'
        , 'Large print', 'reissue'
        , ' a ', 'ss ', 's '
        ]
    l_replace_other = [item.lower() for item in l_replace_other]
    l_replace = l_replace_general + l_replace_other
    # Function for string replace
    def string_simplify(s):
        s = s.lower()
        for item in l_replace:
            s = s.replace(item, '')
        s = re.sub("\(.*\)", '', s)
        s = re.sub('ss$', '', s) # This weird bit helps handle apostrophe s combinations that weren't already handled
        s = re.sub('s$', '', s)
        s = s.replace(' ', '').translate(str.maketrans('', '', string.punctuation))
        return s
    # Initial setup for this
    df_books.fillna('Unknown', inplace=True)
    df_books['title_simple'] = df_books['Book-Title'].apply(string_simplify)
    df_books['author_simple'] = df_books['Book-Author'].apply(string_simplify)
    df_books['title_author_key'] = df_books.title_simple + df_books.author_simple
    df_books.sort_values(by='Year-Of-Publication', ascending=False, inplace=True) # arbitrarily keep the first
    # Create a unique set by title-author in order to create a dictionary mapping
    df_books_unique = df_books.drop_duplicates(subset='title_author_key', keep='first')
    d_title_author_isbn = dict(zip(df_books_unique.title_author_key, df_books_unique.ISBN))
    # Map the parent back to all entries in the books df
    df_books['ISBN_parent'] = df_books.title_author_key.apply(lambda x: d_title_author_isbn[x])
    # Now get the child to parent mappings and apply those to the ratings df
    d_isbn_isbnparent = dict(zip(df_books.ISBN, df_books.ISBN_parent))
    df_ratings['ISBN'] = df_ratings.ISBN.apply(lambda x: d_isbn_isbnparent[x])
    # Now clean back up the books df dropping down to one row per parent
    df_books['ISBN'] = df_books.ISBN_parent
    df_books = df_books.drop_duplicates(subset='ISBN', keep='first').reset_index(drop=True)
    df_books.drop(columns=['title_simple', 'author_simple', 'title_author_key', 'ISBN_parent'], inplace=True)

    # Now, regardless of other choices, remove users that have no nonzero ratings
    df_user_max_all = df_ratings.groupby('User-ID')[['Book-Rating']].max().reset_index()
    l_nonzero_users = df_user_max_all.query('`Book-Rating` > 0')['User-ID'].unique().tolist()
    df_ratings = df_ratings.query('`User-ID` in @l_nonzero_users').reset_index(drop=True)

    # If we want to use the implicit ratings, some handling below
    # Basic case, add 1 to keep it from the nonzero group
    if use_implicit == 'yes':
        df_ratings['Book-Rating'] = df_ratings['Book-Rating'] + 1
    # Or replace 0s with the mean user rating
    elif use_implicit == 'as mean':
        df_user_means_nonzero = df_ratings.query('`Book-Rating` > 0')\
            .groupby('User-ID')[['Book-Rating']].mean()\
            .reset_index()
        d_user_mean = dict(zip(df_user_means_nonzero['User-ID'], df_user_means_nonzero['Book-Rating']))
        df_ratings['Book-Rating'] = df_ratings.apply(lambda x:
                                                     d_user_mean[x['User-ID']] if int(x['Book-Rating']) == 0 else x['Book-Rating']
                                                     , axis=1)
    # Or replace with the lower quantile
    elif use_implicit == 'as quantile':
        df_user_means_nonzero = df_ratings.query('`Book-Rating` > 0')\
            .groupby('User-ID')[['Book-Rating']].quantile(0.25)\
            .reset_index()
        d_user_mean = dict(zip(df_user_means_nonzero['User-ID'], df_user_means_nonzero['Book-Rating']))
        df_ratings['Book-Rating'] = df_ratings.apply(lambda x:
                                                     d_user_mean[x['User-ID']] if int(x['Book-Rating']) == 0 else x['Book-Rating']
                                                     , axis=1)     
    # Otherwise leave things alone
    else:
        pass

    # Get unique ISBNs in both dfs (again in case used by code below)
    a_ratings_ISBN = df_ratings.ISBN.unique()
    a_books_ISBN = df_books.ISBN.unique()

    # Create a copy df of ratings without nonzero values
    df_ratings_nonzero = df_ratings.query('`Book-Rating` > 0').reset_index(drop=True).copy()

    # Now determine which books to keep based on how many times they show up in the ratings
    s_books_values = df_ratings_nonzero.ISBN.value_counts()
    l_books_keep = s_books_values[s_books_values > n_ratings_to_keep_books].index.to_list()
    print(f'You are keeping a total of {len(l_books_keep)} books out of {len(s_books_values)}.')

    # Now remove books with not enough ratings from our dfs
    df_ratings = df_ratings.query('ISBN in @l_books_keep').reset_index(drop=True)
    df_ratings_nonzero = df_ratings_nonzero.query('ISBN in @l_books_keep').reset_index(drop=True)

    # Similarly, determine which users to keep based on how many nonzero ratings
    s_user_ratings = df_ratings_nonzero['User-ID'].value_counts()
    l_users_keep = s_user_ratings[s_user_ratings >= n_ratings_to_keep_users].index.to_list()
    print(f'You are keeping a total of {len(l_users_keep)} users out of {len(s_user_ratings)}.')

    # Remove users accordingly
    df_ratings = df_ratings.query('`User-ID` in @l_users_keep').reset_index(drop=True)
    df_ratings_nonzero = df_ratings_nonzero.query('`User-ID` in @l_users_keep').reset_index(drop=True)

    # Now downfilter the books dataframe to only books left with ratings
    l_books_remaining = df_ratings_nonzero.ISBN.unique().tolist()
    df_books = df_books.query('ISBN in @l_books_remaining').reset_index(drop=True)

    # Because we are using the parent ISBN in the ratings, it's now possible to have duplicate ratings
    # In these cases (i.e. n children on one parent) we will take the mean, so grouping the ratings works
    df_ratings = df_ratings.groupby(['User-ID', 'ISBN'])[['Book-Rating']].mean().reset_index()
    df_ratings_nonzero = df_ratings_nonzero.groupby(['User-ID', 'ISBN'])[['Book-Rating']].mean().reset_index()

    return df_ratings_nonzero, df_books




# %%
# The actual class for the book recommender
class BookRecommender():

    def __init__(self):

        self.d_isbn_idx = None
        self.d_idx_isbn = None
        self.d_isbn_title = None
        self.df_books = None
        self.d_isbn_count = None
        self.embeddings = None
        self.dot_prod = None

    def store_metadata(self, df_matrix, df_books, df_ratings):

        # Get list of isbns and corresponding indices
        l_isbns = df_matrix.index.to_list()
        a_isbn_idx = np.arange(0, len(l_isbns))
        # Merge these together into dicts
        d_isbn_idx = dict(zip(l_isbns, a_isbn_idx))
        d_idx_isbn = dict(zip(a_isbn_idx, l_isbns))
        # Store in the class
        self.d_isbn_idx = d_isbn_idx
        self.d_idx_isbn = d_idx_isbn

        # Now create isbn to title map from books metadata
        d_isbn_title = dict(zip(df_books['ISBN'], df_books['Book-Title']))
        self.d_isbn_title = d_isbn_title

        # Easiest to also just store the books df
        self.df_books = df_books

        # We also want the counts out of df_ratings per ISBN
        # This helps (later) to identify which isbn (of a title) is more common
        s_isbn_counts = df_ratings.ISBN.value_counts()
        self.d_isbn_count = dict(zip(
            s_isbn_counts.index.tolist()
            , s_isbn_counts.values
            ))

    def store_embeddings(self, embeddings):

        self.embeddings = embeddings

    def store_dotprod_function(self, func):

        self.dot_prod = func

    def get_dotprod(self, a1, a2):

        return self.dot_prod(a1, a2)

    def find_isbn_from_title(self, title):

        # Looks up and shows the most similar titles and isbns from a given string
        l_titles = list(self.d_isbn_title.values())
        query_results = process.extractBests(query=title, choices=l_titles, limit=5, scorer=fuzz.token_sort_ratio)
        l_matches = [result[0] for result in query_results]
        l_scores = [result[1] for result in query_results]
        d_match_scores = dict(zip(l_matches, l_scores))
        df_matches = self.df_books.query('`Book-Title` in @l_matches').copy()
        df_matches['match_score'] = df_matches['Book-Title'].apply(lambda x: d_match_scores[x])
        df_matches['ratings_count'] = df_matches['ISBN'].apply(lambda x: self.d_isbn_count[x])
        display(df_matches.sort_values(by='match_score', ascending=False))
    
    def recommend_books(self, isbn, top_n=5):

        # Run the recommendation
        querry_idx = self.d_isbn_idx[isbn]
        querry_ebmedding = self.embeddings[querry_idx]
        a_similarity_scores = self.get_dotprod(self.embeddings, querry_ebmedding)
        a_top_scores = np.sort(a_similarity_scores)[::-1][1:top_n+1]
        a_similar_idxs = a_similarity_scores.argsort()[::-1][1:top_n+1]
        l_similar_isbns = [self.d_idx_isbn[isbn_idx] for isbn_idx in a_similar_idxs]
        d_isbn_score = dict(zip(l_similar_isbns, a_top_scores))

        # Now sculpt the output as a convenient df
        df_similar = self.df_books.query('ISBN in @l_similar_isbns').copy()
        df_similar['similarity_score'] = df_similar.ISBN.apply(lambda x: d_isbn_score[x])
        df_similar.sort_values(by='similarity_score', ascending=False, inplace=True)

        return df_similar


# %%
import pandas as pd

# %%
l_cols = [
    'age'
    ,'workclass'
    ,'fnlwgt'
    ,'education'
    ,'education_num'
    ,'marital_status'
    ,'occupation'
    ,'relationship'
    ,'race'
    ,'male_female'
    ,'capital_gain'
    ,'capital_loss'
    ,'hours_per_week'
    ,'native_country'
    ,'income_group'
    ]
df = pd.read_csv('https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data', index_col=False, names=l_cols)
df.head()

# %%
df['fnlwgt'] = (df.fnlwgt / df.fnlwgt.min()).round()

# %%
df = df.loc[df.index.repeat(df.fnlwgt)].reset_index(drop=True)
print(len(df))
df.head()

# %%
df.drop(columns='fnlwgt', inplace=True)

# %%
df.to_csv('uci_census_data.csv', index=False)

# %%
ratio_fm = len(df.query('male_female == " Male"')) / len(df.query('male_female == " Female"')) 
df['fm'] = df.male_female.apply(lambda x: ratio_fm if x == " Female" else 1)
df = df.loc[df.index.repeat(df.fm)].reset_index(drop=True)
df.drop(columns=['fm'], inplace=True)
diff_fm = len(df.query('male_female == " Male"')) - len(df.query('male_female == " Female"'))
df_fsamp = df.query('male_female == " Female"').sample(diff_fm)
df = pd.concat([df, df_fsamp]).reset_index(drop=True)
print(len(df))

# %%
df.to_csv('uci_census_data_balanced.csv', index=False)

# %%
df.head()


# %%
# Likely needed imports
import numpy as np




# First question check
def elements_check_1(func):
    assert func('sky') == 3, 'Error on sky input'
    assert func('life') == 1, 'Error on life input'
    assert func('energy') == 1, 'Error on energy input'
    assert func('spirit') == 1, 'Error on spirit input'
    print('Correct, good job!')




# Second question check
def elements_check_2(func):
    assert func('sky', 'sky') == 3, 'Error on sky input' # Sky always gets the bonus
    assert func('sky', 'life') == 3, 'Error on sky input'
    assert func('sky', 'energy') == 3, 'Error on sky input'
    assert func('sky', 'spirit') == 3, 'Error on sky input'
    assert func('life', 'sky') == 1, 'Error on life input'
    assert func('life', 'life') == 1, 'Error on life input'
    assert func('life', 'energy') == 1, 'Error on life input'
    assert func('life', 'spirit') == 2, 'Error on life input' # Life against spirit
    assert func('energy', 'sky') == 1, 'Error on energy input'
    assert func('energy', 'life') == 2, 'Error on energy input' # Energy against life
    assert func('energy', 'energy') == 1, 'Error on energy input'
    assert func('energy', 'spirit') == 1, 'Error on energy input'
    assert func('spirit', 'sky') == 1, 'Error on spirit input'
    assert func('spirit', 'life') == 1, 'Error on spirit input'
    assert func('spirit', 'energy') == 2, 'Error on spirit input' # Spirit against energy
    assert func('spirit', 'spirit') == 1, 'Error on spirit input'
    print('Correct, good job!')




# Third question check
def elements_check_3(func):

    # The correct damage calculation
    def correct_calc(your_element, opponent_element, base_damage=3.145):

        if your_element == 'sky':
            dmg_mult = 3
        elif (your_element == 'spirit') and (opponent_element == 'energy'):
            dmg_mult = 2
        elif (your_element == 'energy') and (opponent_element == 'life'):
            dmg_mult = 2
        elif (your_element == 'life') and (opponent_element == 'spirit'):
            dmg_mult = 2
        else:
            dmg_mult = 1

        final_damage = base_damage * dmg_mult
            
        return round(final_damage, 1)

    # Iterate over all the element combos, asserting using the same number (should get all meaningful mistakes)
    l_e = ['sky', 'life', 'energy', 'spirit']
    for your_e in l_e:
        for opponent_e in l_e:
            assert func(your_e, opponent_e) == correct_calc(your_e, opponent_e, 3.145), f'Error on {your_e} vs {opponent_e}.'

    print('Correct, good job!')



# %%
import pandas as pd

l_cols = ["symboling","normalized-losses","make","fuel-type","aspiration", "num-of-doors","body-style",
         "drive-wheels","engine-location","wheel-base", "length","width","height","curb-weight","engine-type",
         "num-of-cylinders", "engine-size","fuel-system","bore","stroke","compression-ratio","horsepower",
         "peak-rpm","city-mpg","highway-mpg","price"]
df = pd.read_csv('https://archive.ics.uci.edu/ml/machine-learning-databases/autos/imports-85.data'
                 , names=l_cols
                 , na_values='?')
df.to_csv('cars.csv', index=False)
df.head()


# %%
# Likely imports for all functions below
import numpy as np
import matplotlib.pyplot as plt
import time



# %%
# Example Likert distributions with various centers and outlier capability
# Returns an array of ints of a given size
def ExLikertDist(size=1000, center=None, outlier=False):

    l_choices = [1, 2, 3, 4, 5]

    if center == 'left':
        l_probs = [0.40, 0.35, 0.15, 0.05, 0.05]
    elif center == 'right':
        l_probs = [0.05, 0.05, 0.15, 0.35, 0.40]
    elif center == 'bimodal':
        l_probs = [0.35, 0.125, 0.05, 0.125, 0.35]
    elif center == 'center':
        l_probs = [0.05, 0.15, 0.60, 0.15, 0.05]
    else:
        l_probs = [0.2, 0.2, 0.2, 0.2, 0.2]

    a = np.random.choice(l_choices, p=l_probs, size=size)

    if outlier:
        a = np.concatenate([a, [8]*int(round(size/100))])

    return a




# %%
# Magic eight ball class for object-oriented activity
class MagicEightBall():

    def __init__(self):

        self.l_choices = [
            'As I see it, yes'
            , 'Ask again later'
            , 'Better not tell you now'
            , 'Cannot predict now'
            , 'Concentrate and ask again'
            , 'Dont count on it'
            , 'It is certain'
            , 'It is decidedly so'
            , 'Most likely'
            , 'What do you think'
            , 'My reply is no'
            , 'My sources say no'
            , 'Why on Earth would you trust the answers of a pool ball?'
            , 'Outlook good'
            , 'Outlook not so good'
            , 'Reply hazy try again'
            , 'Leave me alone, I am tired'
            , 'Signs point to yes'
            , 'Very doubtful'
            , 'Without a doubt'
            , 'Yes'
            , 'Yes, definitely'
            , 'You may rely on it'
            , '...I hate this job. Why do I answer questions all day? Geeze... OH! Sorry there. AHEM, "Your answer is yes"'
            ]

        self.shape = 'round'
        self.color = 'black'
        self.number = 8
        self.size = 'Eh probably like a few inches in diameter?'

    def shake(self):
        print('Why are you shaking me?! Ask me a question')
    
    def answer(self, some_string='Does not matter really'):
        print(np.random.choice(self.l_choices))




# %%
# RobTheRobot class used for object-oriented activity
class RobTheRobot():

    def __init__(self, color=None):

        # Rob may wear different things when he is red
        if np.random.choice([1, 2]) == 1:
            self.red_headgear = 'sombrero'
            self.red_bodygear = 'tuxedo'
        else:
            self.red_headgear = 'snorkel'
            self.red_bodygear = 'cape'           

        # Basic setup
        self.l_all_colors = [
            'red'
            , 'blue'
            , 'green'
            , 'yellow'
            , 'purple'
        ]

        self.d_wearables = {
            'red': {'headgear': self.red_headgear, 'bodygear': self.red_bodygear}
            , 'blue': {'headgear': 'headphones', 'bodygear': 'nothing'}
            , 'green': {'headgear': 'nothing', 'bodygear': 'armor'}
            , 'yellow': {'headgear': 'sombrero', 'bodygear': 'nothing'}
            , 'purple': {'headgear': 'headphones', 'bodygear': 'tuxedo'}
            , 'orange': {'headgear': 'snorkel', 'bodygear': 'cape'}
        }

        # Allows students to specifically pass a color, if not randomly pick one and finish setup
        if color:
            self.color = color
        else:
            self.color = np.random.choice(self.l_all_colors)

        self.headgear = self.d_wearables[self.color]['headgear']
        self.bodygear = self.d_wearables[self.color]['bodygear']

    # Students should provide a function with two arguments (headgear and bodygear) that can guess Rob's color
    # This function checks if the provided function is correct
    def check_answer(self, myfunc=None):

        correct = False

        if myfunc:

            for check_color in self.l_all_colors:
                check_headgear = self.d_wearables[check_color]['headgear']
                check_bodygear = self.d_wearables[check_color]['bodygear']
                answer_color = myfunc(check_headgear, check_bodygear)
                if answer_color != check_color:
                    print(f'Sorry, something went wrong on the color {check_color}.')
                    break
                elif check_color == self.l_all_colors[-1]:
                    correct = True

            if correct:
                print('Awesome, you did it!')

        else:
            print('Provide a function to check your answer!')

        # Note a simple reasonable answer might look like the below
        # def find_robot_color(headgear, bodygear):

        #     if headgear == 'nothing':
        #         color = 'green'
        #     elif bodygear == 'nothing' and headgear == 'headphones':
        #         color = 'blue'
        #     elif bodygear == 'tuxedo' and headgear == 'headphones':
        #         color = 'purple'
        #     elif bodygear == 'nothing' and headgear == 'sombrero':
        #         color = 'yellow'
        #     else:
        #         color = 'red'
            
        #     return color




# %%
# Timer for the function creation activity
def time_me():
    start_t = time.time()
    while True:
        time.sleep(1)
        elapsed_seconds = round(time.time() - start_t)
        print('\r' + 'Seconds elapsed: ' + str(elapsed_seconds), end='')
        if elapsed_seconds >= 45:
            print()
            break




# %%
# Check answer function for the pandas sphinx riddle
def riddle_of_the_sphinxes(answer):
    if answer == 3678:
        print('You grab the sphinx maked 640 and inspect it carefully. Upon flipping it over, you notice a secret compartment.')
        print('Inside is a small note with lattitude and longitude coordinates. Congrats! You are about to be rich.')
        print('\n')
        print('You have solved the riddle! For bonus points, come up with your own riddle! But you must show the work to prove that it works!')
    else:
        print('Sorry, no treasure here!')


# %%
import numpy as np
import pandas as pd
np.random.seed(42)


# %%
n = 10000

d_material = {
    'clay': 0.1
    , 'sandstone': 0.15
    , 'porcelain': 0.2
    , 'gold': 0.05
    , 'iron': 0.1
    , 'copper': 0.15
    , 'bronze': 0.25
    }
d_condition = {
    'poor': 0.25
    , 'fair': 0.35
    , 'good': 0.3
    , 'excellent': 0.1
}
d_paint_coloring = {
    'absent': 0.60
    , 'mostly blue': 0.04
    , 'mostly green': 0.05
    , 'mostly red': 0.01
    , 'various': 0.30
}
d_posture = {
    'crouched': 0.25
    , 'pouncing': 0.15
    , 'standing': 0.15
    , 'laying': 0.40
    , 'other': 0.05
}
d_markings = {
    'plain': 0.3
    , 'light': 0.2
    , 'medium': 0.4
    , 'heavy': 0.1
}

a_id = np.arange(1, n+1)

a_material = np.random.choice(list(d_material.keys()), size=n, p=list(d_material.values()))
a_condition = np.random.choice(list(d_condition.keys()), size=n, p=list(d_condition.values()))
a_paint_coloring = np.random.choice(list(d_paint_coloring.keys()), size=n, p=list(d_paint_coloring.values()))
a_posture = np.random.choice(list(d_posture.keys()), size=n, p=list(d_posture.values()))
a_markings = np.random.choice(list(d_markings.keys()), size=n, p=list(d_markings.values()))
a_headdress_present = np.random.choice(['Y', 'N'], size=n, p=[0.7, 0.3])
a_has_wings = np.random.choice([True, False], size=n, p=[0.15, 0.85])
a_has_tail = np.random.choice([True, False], size=n, p=[0.40, 0.60])

a_paw_digits = np.random.choice([2, 3, 4, 5], size=n, p=[0.2, 0.1, 0.4, 0.3])

a_length_cm_1 = np.random.normal(loc=(6*2.54), scale=1.0, size=5000)
a_length_cm_2 = np.random.normal(loc=(15*2.54), scale=2.0, size=3000)
a_length_cm_3 = np.random.normal(loc=(30*2.54), scale=3.0, size=1500)
a_length_cm_4 = np.random.normal(loc=(90*2.54), scale=4.0, size=500)
a_length = np.concatenate([a_length_cm_1, a_length_cm_2, a_length_cm_3, a_length_cm_4])

df_sphinx = pd.DataFrame({
    'id':a_id
    , 'material':a_material
    , 'condition':a_condition
    , 'paint_coloring':a_paint_coloring
    , 'posture':a_posture
    , 'markings':a_markings
    , 'headdress_present':a_headdress_present
    , 'has_wings':a_has_wings
    , 'has_tail':a_has_tail
    , 'number_paw_digits':a_paw_digits
    , 'length_cm':a_length

})

df_sphinx.head()

# %%
def get_width(length):
    l_fracs = [0.3, 0.4, 0.5]
    frac = np.random.choice(l_fracs)
    return length * frac

df_sphinx['width_cm'] = df_sphinx.length_cm.apply(get_width)

def get_height(width):
    l_fracs = [0.80, 0.9, 1.05, 1.1, 1.2, 1.3, 1.4, 1.5]
    frac = np.random.choice(l_fracs)
    return width * frac

df_sphinx['height_cm'] = df_sphinx.width_cm.apply(get_height)

df_sphinx.head()

# %%
d_density_g_cm3 = {
    'clay': 2.65
    , 'sandstone': 2.3
    , 'porcelain': 2.403
    , 'gold': 19.32
    , 'iron': 7.874
    , 'copper': 8.96
    , 'bronze': 8.73
}

def get_density(material):
    mat_density = d_density_g_cm3[material]
    this_density = np.random.normal(size=1, loc=mat_density, scale=0.1 * mat_density)[0]
    return this_density

df_sphinx['density_g_cm3'] = df_sphinx.material.apply(get_density)
df_sphinx.sample(5)

# %%
df_sphinx['age_years'] = (4500 - (4500 * np.random.beta(a=20, b=2, size=n))).round()
df_sphinx['age_years'] = df_sphinx.age_years.astype('int')
df_sphinx.age_years.describe()

# %%
df_sphinx.sample(10)

# %%
df_sphinx.iloc[327, 5] = np.nan

# %%
df_sphinx.query('markings != markings')

# %%
df_sphinx.to_csv('df_sphinx.csv', index=False)
